/*********************************************************************
*                         F I L A B U I L D                          *
**********************************************************************
* Modified version of Situs hlxbuild.c (c) Willy Wriggers, 1998-2013 *
* Specialized tool to create a number of helical filaments           *
**********************************************************************
*                                                                    * 
* compile on Unix platforms:                                         *
*       cc -O -c stdread.c                                           *
*       cc -O -c volio.c                                             *
*       cc -O -c pdbio.c                                             *
*       cc -O -c filabuild.c                                         *
*       cc -O -o filabuild filabuild.o volio.o stdread.o pdbio.o -lm *
*                                                                    *
**********************************************************************
* See legal statement for terms of distribution                      *
*********************************************************************/


#include "situs.h"

#define NFILMAX 20  /* maximum number of filaments */

/* function declarations */

extern void readpdb(char *, int *, PDB **);
extern void writepdb(char *, int, PDB *);
extern void readvol(char *, MYFLT *, MYFLT *, MYFLT *, MYFLT *, int *, int *, int *, MYFLT **);
extern int readlnint();
extern MYFLT readlnmyflt();

main(int argc, char *argv[])

{ int itot, ibeg[NFILMAX], iend[NFILMAX];
  MYFLT theta[NFILMAX], zrise[NFILMAX];
  MYFLT xoff[NFILMAX], yoff[NFILMAX];
  MYFLT thetaoff[NFILMAX];
  int flipflag[NFILMAX];
  int numAtoms, i, j, k, kold, l, unit;
  PDB *pdbP, *pdbH;
  MYFLT vcos, vsin;
  double dtemp;
  char newsegid[5];
  int cfil, nfil;
  Rseq3 com1, com2;
  MYFLT xsum, xdif, ysum, ydif, xold, yold, zold, norm1, norm2;
  FILE *fin;

  if (argc < 3 || argc > 4) {
    fprintf(stderr, "filabuild> Usage: filabuild input(PDB)-file [optional: input(configuration)-file] output(PDB)-file \n");
    exit (1);
  }
  
  readpdb(argv[1], &numAtoms, &pdbP);

  if (argc == 4) { /* read configuration file */

    printf("filabuild> Processing configuration file %s \n",argv[2]); 
    fin = fopen(argv[2], "r"); 
    if(fin == NULL) {
      fprintf(stderr, "filabuild> Error: Can't open file %s [e.c. 70000]\n",argv[2]); 
      exit(70000);   
    }
    nfil = 0;
    if (fscanf (fin, "%d", &nfil) != 1) {
      fprintf(stderr, "filabuild> Error: Can't read number of filaments in file %s [e.c. 70005]\n",argv[2]); 
      exit(70005);
    }
    if (nfil < 0) {
      fprintf(stderr, "filabuild> Error: Number of filaments in file %s must be positive [e.c. 70010]\n",argv[2]); 
      exit(70010);
    }
    if (nfil > NFILMAX) {
      fprintf(stderr, "filabuild> Error: Number of filaments in file %s must be smaller than NFILMAX [e.c. 70015]\n",argv[2]);
      fprintf(stderr, "filabuild> Increase NFILMAX in filabuild.c or adjust configuration.\n");
      exit(70015);
    }
    for(cfil=0;cfil<nfil;++cfil) {
      if (fscanf (fin, "%le", &dtemp) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read zrise[%d] in file %s [e.c. 70020]\n", cfil, argv[2]); 
	exit(70020);
      }
      zrise[cfil]=dtemp;
      if (fscanf (fin, "%le", &dtemp) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read theta[%d] in file %s [e.c. 70025]\n", cfil, argv[2]); 
	exit(70025);
      }
      theta[cfil]=dtemp;
      if (fscanf (fin, "%le", &dtemp) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read thetaoff[%d] in file %s [e.c. 70030]\n", cfil, argv[2]); 
	exit(70030);
      }
      thetaoff[cfil]=dtemp;
      if (fscanf (fin, "%d", (ibeg+cfil)) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read ibeg[%d] in file %s [e.c. 70035]\n", cfil, argv[2]); 
	exit(70035);
      }
      ibeg[cfil] *= -1;
      if (fscanf (fin, "%d", (iend+cfil)) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read iend[%d] in file %s [e.c. 70040]\n", cfil, argv[2]); 
	exit(70040);
      }
      if (iend[cfil]<ibeg[cfil]) {
	fprintf(stderr, "filabuild> Warning: iend[%d]: %d must be >= ibeg[%d]: %d \n", cfil, iend[cfil], cfil, ibeg[cfil]);
	iend[cfil] = ibeg[cfil];
	fprintf(stderr, "filabuild> iend[%d] set to %d \n", cfil, ibeg[cfil]);
      }
      if (fscanf (fin, "%le", &dtemp) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read xoff[%d] in file %s [e.c. 70045]\n", cfil, argv[2]); 
	exit(70045);
      }
      xoff[cfil]=dtemp;
      if (fscanf (fin, "%le", &dtemp) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read yoff[%d] in file %s [e.c. 70050]\n", cfil, argv[2]); 
	exit(70050);
      }
      yoff[cfil]=dtemp;
      if (fscanf (fin, "%d", (flipflag+cfil)) != 1) {
	fprintf(stderr, "filabuild> Error: Can't read flipflag[%d] in file %s [e.c. 70055]\n", cfil, argv[2]); 
	exit(70055);
      }
      if (flipflag[cfil]<0 || flipflag[cfil]>1) {
	fprintf(stderr, "filabuild> Error: flipflag[%d] in file %s out of range [e.c. 70060]\n", cfil, argv[2]); 
	exit(70060);
      }
    }
    fclose(fin);

  } else { /* enter the configuration manually */

    printf ("filabuild> Enter helical and lattice parameters manually: \n");
    printf ("filabuild> Enter number of filaments: ");
    nfil = readlnint();
    if (nfil < 0) {
      fprintf(stderr, "filabuild> Error: Number of filaments in file %s must be positive [e.c. 70110]\n",argv[2]); 
      exit(70110);
    }
    if (nfil > NFILMAX) {
      fprintf(stderr, "filabuild> Error: Number of filaments in file %s must be smaller than NFILMAX [e.c. 70115]\n",argv[2]);
      fprintf(stderr, "filabuild> Increase NFILMAX in filabuild.c or adjust configuration.\n");
      exit(70115);
    }
    for(cfil=0;cfil<nfil;++cfil) {
      printf ("filabuild> \n");
      printf ("filabuild> Filament Nr. %d: \n",cfil+1);
      printf ("filabuild> \n");
      printf ("filabuild> Enter helical rise per monomer in Angstrom: ");
      zrise[cfil] = readlnmyflt();
      printf ("filabuild> Enter angular twist per monomer in degrees: ");  
      theta[cfil] = readlnmyflt();
      printf ("filabuild> Enter twist-offset in degrees : ");   
      thetaoff[cfil] = readlnmyflt();
      printf ("filabuild> Enter number of monomers before input monomer: ");  
      ibeg[cfil] = readlnint(); ibeg[cfil] *= -1;
      printf ("filabuild> Enter number of monomers after input monomer: ");  
      iend[cfil] = readlnint();
      if (iend[cfil]<ibeg[cfil]) {
	fprintf(stderr, "filabuild> Warning: iend[%d]: %d must be >= ibeg[%d]: %d \n", cfil, iend[cfil], cfil, ibeg[cfil]);
	iend[cfil] = ibeg[cfil];
	fprintf(stderr, "filabuild> iend[%d] set to %d \n", cfil, ibeg[cfil]);
      }
      printf ("filabuild> Enter x-offset in Angstrom: ");  
      xoff[cfil] = readlnmyflt();
      printf ("filabuild> Enter y-offset in Angstrom: ");   
      yoff[cfil] = readlnmyflt();
      printf ("filabuild> Enter flag for flip of filament (0=none, 1=flipped): ");   
      flipflag[cfil] = readlnint();
      if (flipflag[cfil]<0 || flipflag[cfil]>1) {
	fprintf(stderr, "filabuild> Error: flipflag out of range [e.c. 70120]\n"); 
	exit(70120);
      }
    }
  }
  printf("filabuild> Parameters for %d filament",nfil);
  if (nfil>1) printf("s");
  printf(" read\n");

  /* allocate output pdb storage */
  itot = 0;
  for(cfil=0;cfil<nfil;++cfil) itot += iend[cfil]-ibeg[cfil]+1;

  pdbH = (PDB *) malloc(numAtoms * sizeof(PDB) * itot);
  if (pdbH == NULL) {
    fprintf(stderr, "filabuild> Error: Could not satisfy memory allocation request [e.c. 70210]\n"); 
    exit(70210);
  }

  k = 0; /* counter for nr. of duplicates */
  for(cfil=0;cfil<nfil;++cfil) {
  printf("filabuild> Creating symmetry-related duplicates for filament nr. %d.\n",cfil+1);
    for(unit=ibeg[cfil]; unit<=iend[cfil]; unit++) {
      for(i=0; i<numAtoms; i++) {
        /* transform coordinates */ 
        vcos = cos(PI*((MYFLT)unit*theta[cfil]+thetaoff[cfil])/180.0);
        vsin = sin(PI*((MYFLT)unit*theta[cfil]+thetaoff[cfil])/180.0);
        pdbH[i+numAtoms*k].x = vcos * pdbP[i].x - vsin * pdbP[i].y;   
        pdbH[i+numAtoms*k].y = vsin * pdbP[i].x + vcos * pdbP[i].y;  
        pdbH[i+numAtoms*k].z = pdbP[i].z+(MYFLT)unit*zrise[cfil]; 
	
        /* add number to segid, keeping two letters, if possible */
        sprintf(newsegid, "%d\0", k+1);
        j=0;
	while (newsegid[j] != '\0' && j<4) {
	   pdbH[i+numAtoms*k].segID[j] = newsegid[j];
           ++j;
        }
        l = 0;
        while (pdbP[i].segID[l] != '\0' && j<4 && l<2) {
          pdbH[i+numAtoms*k].segID[j] = pdbP[i].segID[l];
          ++j; ++l;
	}
	pdbH[i+numAtoms*k].segID[j] = '\0';

        /* update atom number */
	pdbH[i+numAtoms*k].serial = i+numAtoms*k;

        /* keep rest as is */
        for (j=0;j<7;++j) pdbH[i+numAtoms*k].recdName[j] = pdbP[i].recdName[j];
        for (j=0;j<3;++j) pdbH[i+numAtoms*k].atomType[j] = pdbP[i].atomType[j];
        for (j=0;j<3;++j) pdbH[i+numAtoms*k].atomLoc[j] = pdbP[i].atomLoc[j];
        for (j=0;j<2;++j) pdbH[i+numAtoms*k].altLoc[j] = pdbP[i].altLoc[j];
        for (j=0;j<5;++j) pdbH[i+numAtoms*k].resName[j] = pdbP[i].resName[j]; 
        for (j=0;j<2;++j) pdbH[i+numAtoms*k].chainID[j] = pdbP[i].chainID[j];
        pdbH[i+numAtoms*k].resSeq = pdbP[i].resSeq;
        for (j=0;j<2;++j) pdbH[i+numAtoms*k].iCode[j] = pdbP[i].iCode[j]; 
        pdbH[i+numAtoms*k].occupancy = pdbP[i].occupancy;   
        pdbH[i+numAtoms*k].tempFactor = pdbP[i].tempFactor;
	pdbH[i+numAtoms*k].ftNote = pdbP[i].ftNote;  
        for (j=0;j<3;++j) pdbH[i+numAtoms*k].element[j] = pdbP[i].element[j];
        for (j=0;j<3;++j) pdbH[i+numAtoms*k].charge[j] = pdbP[i].charge[j];
      }
      k++;
    }

    kold = k-1-(iend[cfil]-ibeg[cfil]);

    /* flip coordinates if necessary                                */
    /* the flipaxis is chosen such that the geometric center of the */ 
    /* last monomer coincides with the previous center of the first */
    /* monomer (this preserves the appearance of the helix)         */
    
    if (flipflag[cfil]==1) {

      /* compute geometric center of the first monomer */ 
      for(i=0;i<3;i++) com1[i]=0; 
      for(i=0; i<numAtoms; i++) { 
	com1[0]+=pdbH[i+numAtoms*kold].x; 
	com1[1]+=pdbH[i+numAtoms*kold].y; 
	com1[2]+=pdbH[i+numAtoms*kold].z;
      }
      for(i=0;i<3;i++) com1[i]/=(1.0*numAtoms);
      
      /* compute geometric center of the last monomer */ 
      for(i=0;i<3;i++) com2[i]=0; 
      for(i=0; i<numAtoms; i++) { 
	com2[0]+=pdbH[i+numAtoms*(k-1)].x; 
	com2[1]+=pdbH[i+numAtoms*(k-1)].y; 
	com2[2]+=pdbH[i+numAtoms*(k-1)].z;
      }
      for(i=0;i<3;i++) com2[i]/=(1.0*numAtoms);
   
      /* assign flip axis coordinate system */
      xsum = com1[0]+com2[0];
      xdif = com1[0]-com2[0];
      ysum = com1[1]+com2[1];
      ydif = com1[1]-com2[1];
      norm1 = xsum*xsum+ysum*ysum;
      norm2 = xdif*xdif+ydif*ydif;

      /* handle special cases of parallel and orthogonal com vectors */
      if (norm2 < 0.01*norm1) {   /* parallel */
	xdif = -ysum;
	ydif = xsum;
	norm2 = xdif*xdif+ydif*ydif;
      } else if (norm1 < 0.01*norm2) {   /* orthogonal */
	xsum = -ydif;
	ysum = xdif;
	norm1 = xsum*xsum+ysum*ysum;
      }

      for(i=0; i<numAtoms*(k-kold); i++) {
	xold = pdbH[i+numAtoms*kold].x;
	yold = pdbH[i+numAtoms*kold].y;
	zold = pdbH[i+numAtoms*kold].z;
        pdbH[i+numAtoms*kold].x = (xsum*xsum*xold + ysum*xsum*yold)/norm1 -    
	                          (xdif*xdif*xold + ydif*xdif*yold)/norm2 ;
        pdbH[i+numAtoms*kold].y = (xsum*ysum*xold + ysum*ysum*yold)/norm1 -    
	                          (xdif*ydif*xold + ydif*ydif*yold)/norm2 ;
        pdbH[i+numAtoms*kold].z = com1[2] + com2[2] - zold; 
      }
    }

    /* apply (x,y)-offset of filament */
    for(i=0; i<numAtoms*(k-kold); i++) {
      pdbH[i+numAtoms*kold].x += xoff[cfil];
      pdbH[i+numAtoms*kold].y += yoff[cfil];
    }
  }

  writepdb(argv[argc-1], numAtoms*k, pdbH);
  printf ("filabuild> %d atoms written to file %s.\n",numAtoms*k,argv[argc-1]); 
} 







